import copy
from collections.abc import MutableSequence

import networkx as nx
import sympy

import pharmpy.symbols as symbols
import pharmpy.unicode as unicode


class Assignment:
    """Representation of variable assignment

    This class is similar to :class:`sympy.codegen.Assignment` and are
    combined together into a ModelStatements object.

    Attributes
    ----------
    symbol : sympy.Symbol
        Symbol of statement
    expression : sympy.Expr
        Expression of statement
    """

    def __init__(self, symbol, expression):
        try:
            symbol.is_Symbol
            self.symbol = symbol
        except AttributeError:
            self.symbol = sympy.Symbol(symbol)
        self.expression = sympy.sympify(expression)

    def subs(self, substitutions):
        """Substitute expressions or symbols in assignment

        Parameters
        ----------
        substitutions : dict
            old-new pairs

        Examples
        --------
        >>> from pharmpy import Assignment
        >>> a = Assignment('CL', 'POP_CL + ETA_CL')
        >>> a
        CL := ETA_CL + POP_CL
        >>> a.subs({'ETA_CL' : 'ETA_CL * WGT'})
        >>> a
        CL := ETA_CL⋅WGT + POP_CL

        """
        self.symbol = self.symbol.subs(substitutions, simultaneous=True)
        self.expression = self.expression.subs(substitutions, simultaneous=True)

    @property
    def free_symbols(self):
        """Get set of all free symbols in the assignment

        Note that the left hand side symbol will be in the set

        Examples
        --------
        >>> from pharmpy import Assignment
        >>> a = Assignment('CL', 'POP_CL + ETA_CL')
        >>> a.free_symbols      # doctest: +SKIP
        {CL, ETA_CL, POP_CL}

        """
        symbols = {self.symbol}
        symbols |= self.expression.free_symbols
        return symbols

    @property
    def rhs_symbols(self):
        """Get set of all free symbols in the right hand side expression

        Examples
        --------
        >>> from pharmpy import Assignment
        >>> a = Assignment('CL', 'POP_CL + ETA_CL')
        >>> a.rhs_symbols      # doctest: +SKIP
        {ETA_CL, POP_CL}

        """
        return self.expression.free_symbols

    def __eq__(self, other):
        return (
            isinstance(other, Assignment)
            and self.symbol == other.symbol
            and self.expression == other.expression
        )

    def __repr__(self):
        expression = sympy.pretty(self.expression)
        lines = [line.rstrip() for line in expression.split('\n')]
        definition = f'{sympy.pretty(self.symbol)} := '
        s = ''
        for line in lines:
            if line == lines[-1]:
                s += definition + line + '\n'
            else:
                s += len(definition) * ' ' + line + '\n'
        return s.rstrip()

    def __deepcopy__(self, memo):
        return type(self)(self.symbol, self.expression)

    def copy(self):
        """Create a copy of the Assignment object"""
        return copy.deepcopy(self)

    def _repr_latex_(self):
        sym = self.symbol._repr_latex_()[1:-1]
        expr = self.expression._repr_latex_()[1:-1]
        return f'${sym} := {expr}$'


class ODESystem:
    """Base class and placeholder for ODE systems of different forms

    Attributes
    ----------
    solver : str
        Solver to use when numerically solving the ode system
        Supported solvers and their NONMEM ADVAN

        +------------------------+------------------+
        | Solver                 | NONMEM ADVAN     |
        +========================+==================+
        | CVODES                 | ADVAN14          |
        +------------------------+------------------+
        | DGEAR                  | ADVAN8           |
        +------------------------+------------------+
        | DVERK                  | ADVAN6           |
        +------------------------+------------------+
        | IDA                    | ADVAN15          |
        +------------------------+------------------+
        | LSODA                  | ADVAN13          |
        +------------------------+------------------+
        | LSODI                  | ADVAN9           |
        +------------------------+------------------+
    """

    t = symbols.symbol('t')

    def __init__(self):
        self._solver = None

    @property
    def free_symbols(self):
        return set()

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, value):
        supported = ['LSODA']
        if not (value is None or value.upper() in supported):
            raise ValueError(f"Unknown solver {value}. Recognized solvers are {supported}.")
        self._solver = value

    @property
    def rhs_symbols(self):
        return set()

    def subs(self, substitutions):
        pass

    def __eq__(self, other):
        return isinstance(other, ODESystem)

    def __repr__(self):
        return 'ODESystem()'

    def _repr_html_(self):
        return str(self)

    def _repr_latex_(self):
        return str(self)


def _bracket(a):
    """Append a left bracket for an array of lines"""
    if len(a) == 1:
        return '{' + a[0]
    if len(a) == 2:
        a.append('')
    if (len(a) % 2) == 0:
        upper = len(a) // 2 - 1
    else:
        upper = len(a) // 2
    a[0] = '⎧' + a[0]
    for i in range(1, upper):
        a[i] = '⎪' + a[i]
    a[upper] = '⎨' + a[upper]
    for i in range(upper + 1, len(a) - 1):
        a[i] = '⎪' + a[i]
    a[-1] = '⎩' + a[-1]
    return '\n'.join(a) + '\n'


class ExplicitODESystem(ODESystem):
    """System of ODEs described explicitly

    Attributes
    ----------
    odes : list
        Symbolic differential equations
    ics : dict
        Symbolic initial conditions

    Examples
    --------
    >>> from pharmpy import ExplicitODESystem
    >>> import sympy
    >>> A_DEPOT = sympy.Function('A_DEPOT')
    >>> A_CENTRAL = sympy.Function('A_CENTRAL')
    >>> t = ExplicitODESystem.t
    >>> AMT, KA, K = sympy.symbols('AMT KA K')
    >>> eq1 = sympy.Eq(sympy.Derivative(A_DEPOT(t), t), -KA * A_DEPOT(t))
    >>> eq2 = sympy.Eq(sympy.Derivative(A_CENTRAL(t)), -K*A_CENTRAL(t) + KA*A_DEPOT(t))
    >>> ics = {A_DEPOT(0): AMT, A_CENTRAL(0): 0}
    >>> odes = ExplicitODESystem([eq1, eq2], ics)
    >>> odes
    ⎧d
    ⎪──(A_DEPOT(t)) = -KA⋅A_DEPOT(t)
    ⎪dt
    ⎨d
    ⎪──(A_CENTRAL(t)) = -K⋅A_CENTRAL(t) + KA⋅A_DEPOT(t)
    ⎪dt
    ⎪A_DEPOT(0) = AMT
    ⎩A_CENTRAL(0) = 0
    """

    def __init__(self, odes, ics):
        self.odes = odes
        self.ics = ics
        super().__init__()

    @property
    def free_symbols(self):
        """Get set of all free symbols in the ODE system

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.to_explicit_system()
        >>> model.statements.ode_system.free_symbols  # doctest: +SKIP
        {AMT, CL, V, t}
        """

        free = set()
        for ode in self.odes:
            free |= ode.free_symbols
        for key, value in self.ics.items():
            free |= key.free_symbols
            try:  # To allow for regular python classes as values for ics
                free |= value.free_symbols
            except AttributeError:
                pass
        return free

    def subs(self, substitutions):
        """Substitute expressions or symbols in ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.to_explicit_system()
        >>> model.statements.ode_system.subs({'AMT': 'DOSE'})
        >>> model.statements.ode_system
        ⎧d                  -CL⋅A_CENTRAL(t)
        ⎪──(A_CENTRAL(t)) = ─────────────────
        ⎪dt                         V
        ⎨d                 CL⋅A_CENTRAL(t)
        ⎪──(A_OUTPUT(t)) = ───────────────
        ⎪dt                       V
        ⎪A_CENTRAL(0) = DOSE
        ⎩A_OUTPUT(0) = 0
        """
        d = {
            sympy.Function(str(key))(symbols.symbol('t')): value
            for key, value in substitutions.items()
        }
        d.update(substitutions)
        self.odes = [ode.subs(d) for ode in self.odes]
        self.ics = {key.subs(d): value.subs(d) for key, value in self.ics.items()}

    @property
    def rhs_symbols(self):
        """Get set of all free symbols in the right hand side expressions

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.to_explicit_system()
        >>> model.statements.ode_system.rhs_symbols   # doctest: +SKIP
        {AMT, CL, V, t}
        """
        return self.free_symbols

    def __repr__(self):
        a = []
        for ode in self.odes:
            ode_str = sympy.pretty(ode)
            a += ode_str.split('\n')
        for key, value in self.ics.items():
            ics_str = sympy.pretty(sympy.Eq(key, value))
            a += ics_str.split('\n')
        return _bracket(a)

    def __deepcopy__(self, memo):
        newone = type(self)(copy.copy(self.odes), copy.copy(self.ics))
        return newone

    def __eq__(self, other):
        return (
            isinstance(other, ExplicitODESystem)
            and self.odes == other.odes
            and self.ics == other.ics
            and self.solver == other.solver
        )

    def _repr_latex_(self):
        rows = []
        for ode in self.odes:
            ode_repr = ode._repr_latex_()[1:-1]
            rows.append(ode_repr)
        for k, v in self.ics.items():
            ics_eq = sympy.Eq(k, v)
            ics_repr = ics_eq._repr_latex_()[1:-1]
            rows.append(ics_repr)
        return r'\begin{cases} ' + r' \\ '.join(rows) + r' \end{cases}'

    def to_compartmental_system(self):
        """Get the explicit system as a compartmental ODE system

        Results
        -------
        CompartmentalSystem
            The same ODE system in compartmental representation

        Example
        -------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.to_explicit_system()
        >>> odes = model.statements.ode_system.to_compartmental_system()
        >>> odes
        Bolus(AMT)
        ┌───────┐       ┌──────┐
        │CENTRAL│──CL/V→│OUTPUT│
        └───────┘       └──────┘
        """

        def convert_name(name):
            if name.startswith('A_'):
                return name[2:]
            else:
                return name

        funcs = [eq.lhs.args[0] for eq in self.odes]
        cs = CompartmentalSystem()
        for f in funcs:
            cs.add_compartment(convert_name(f.name))

        for eq in self.odes:
            for comp_func in funcs:
                dep = eq.rhs.as_independent(comp_func, as_Add=True)[1]
                if dep == 0:
                    continue
                terms = sympy.Add.make_args(dep)
                for term in terms:
                    expr = term / comp_func
                    if term.args[0] != -1:
                        from_comp = cs.find_compartment(convert_name(comp_func.name))
                        to_comp = cs.find_compartment(convert_name(eq.lhs.args[0].name))
                        cs.add_flow(from_comp, to_comp, expr)

        dose = Bolus("AMT")  # FIXME: not true in general!
        cs.find_compartment(convert_name(funcs[0].name)).dose = dose
        return cs


class CompartmentalSystem(ODESystem):
    """System of ODEs descibed as a compartmental system

    Examples
    --------
    >>> from pharmpy import Bolus, CompartmentalSystem
    >>> cs = CompartmentalSystem()
    >>> dose = Bolus("AMT")
    >>> central = cs.add_compartment("CENTRAL")
    >>> central.dose = dose
    >>> peripheral = cs.add_compartment("PERIPHERAL")
    >>> output = cs.add_compartment("OUTPUT")
    >>> cs.add_flow(central, peripheral, "K12")
    >>> cs.add_flow(peripheral, central, "K21")
    >>> cs.add_flow(central, output, "CL / V")
    >>> cs        # doctest: +SKIP
    Bolus(AMT)
                     ┌──────────┐
                     │PERIPHERAL│
                     └──────────┘
                      ↑        │
                     K12      K21
                      │        ↓
    ┌───────┐      ┌──────────┐      ┌──────────┐       ┌──────┐
    │CENTRAL│──K12→│PERIPHERAL│──K21→│ CENTRAL  │──CL/V→│OUTPUT│
    └───────┘      └──────────┘      └──────────┘       └──────┘
    """

    def __init__(self):
        self._g = nx.DiGraph()
        super().__init__()

    @property
    def free_symbols(self):
        """Get set of all free symbols in the compartmental system

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.free_symbols  # doctest: +SKIP
        {AMT, CL, V, t}
        """
        free = {symbols.symbol('t')}
        for (_, _, rate) in self._g.edges.data('rate'):
            free |= rate.free_symbols
        for node in self._g.nodes:
            free |= node.free_symbols
        return free

    @property
    def rhs_symbols(self):
        """Get set of all free symbols in the right hand side expressions

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.rhs_symbols   # doctest: +SKIP
        {AMT, CL, V, t}
        """
        return self.free_symbols  # This works currently

    def subs(self, substitutions):
        """Substitute expressions or symbols in ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.subs({'AMT': 'DOSE'})
        >>> model.statements.ode_system
        Bolus(DOSE)
        ┌───────┐       ┌──────┐
        │CENTRAL│──CL/V→│OUTPUT│
        └───────┘       └──────┘
        """
        for (u, v, rate) in self._g.edges.data('rate'):
            rate_sub = rate.subs(substitutions, simultaneous=True)
            self._g.edges[u, v]['rate'] = rate_sub
        for comp in self._g.nodes:
            comp.subs(substitutions)

    def atoms(self, cls):
        """Get set of all symbolic atoms of some kind

        For more information see
        https://docs.sympy.org/latest/modules/core.html#sympy.core.basic.Basic.atoms

        Parameters
        ----------
        cls : type
            Type of atoms to find

        Returns
        -------
        set
            Set of symbolic atoms
        """
        atoms = set()
        for (_, _, rate) in self._g.edges.data('rate'):
            atoms |= rate.atoms(cls)
        return atoms

    def __eq__(self, other):
        return (
            isinstance(other, CompartmentalSystem)
            and nx.to_dict_of_dicts(self._g) == nx.to_dict_of_dicts(other._g)
            and self.dosing_compartment.dose == other.dosing_compartment.dose
            and self.solver == other.solver
        )

    def __deepcopy__(self, memo):
        newone = type(self)()
        newone._g = copy.deepcopy(self._g, memo)
        return newone

    def add_compartment(self, name):
        """Create and add compartment to system

        The compartment will be added without any flows to other compartments.
        Use the add_flow method to add flows to and from the newly added compartment.

        Parameters
        ----------
        name : str
            Name for created compartment

        Returns
        -------
        Compartment
            Newly created compartment

        Examples
        --------
        >>> from pharmpy import CompartmentalSystem
        >>> odes = CompartmentalSystem()
        >>> central = odes.add_compartment("CENTRAL")
        """
        comp = Compartment(name)
        self._g.add_node(comp)
        return comp

    def remove_compartment(self, compartment):
        """Remove compartment from system

        Parameters
        ----------
        compartment : Compartment
            Compartment object to remove from system

        Examples
        --------
        >>> from pharmpy import CompartmentalSystem
        >>> odes = CompartmentalSystem()
        >>> central = odes.add_compartment("CENTRAL")
        >>> odes.remove_compartment(central)
        """
        self._g.remove_node(compartment)

    def add_flow(self, source, destination, rate):
        """Add flow between two compartments

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment
        rate : Expression
            Symbolic rate of flow

        Examples
        --------
        >>> from pharmpy import CompartmentalSystem
        >>> odes = CompartmentalSystem()
        >>> depot = odes.add_compartment("DEPOT")
        >>> central = odes.add_compartment("CENTRAL")
        >>> odes.add_flow(depot, central, "KA")
        """
        self._g.add_edge(source, destination, rate=sympy.sympify(rate))

    def remove_flow(self, source, destination):
        """Remove flow between two compartments

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment

        Examples
        --------
        >>> from pharmpy import CompartmentalSystem
        >>> odes = CompartmentalSystem()
        >>> depot = odes.add_compartment("DEPOT")
        >>> central = odes.add_compartment("CENTRAL")
        >>> odes.add_flow(depot, central, "KA")
        >>> odes.remove_flow(depot, central)
        """
        self._g.remove_edge(source, destination)

    def get_flow(self, source, destination):
        """Get the rate of flow between two compartments

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment

        Returns
        -------
        list
            Symbolic rates

        Examples
        --------
        >>> from pharmpy import CompartmentalSystem
        >>> odes = CompartmentalSystem()
        >>> depot = odes.add_compartment("DEPOT")
        >>> central = odes.add_compartment("CENTRAL")
        >>> odes.add_flow(depot, central, "KA")
        >>> odes.get_flow(depot, central)
        KA
        >>> odes.get_flow(central, depot)
        """
        try:
            rate = self._g.edges[source, destination]['rate']
        except KeyError:
            rate = None
        return rate

    def get_compartment_outflows(self, compartment):
        """Get list of all flows going out from a compartment

        Parameters
        ----------
        compartment : Compartment or str
            Get outflows for this compartment

        Returns
        -------
        list
            Pairs of compartments and symbolic rates

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.get_compartment_outflows("CENTRAL")
        [(Compartment(OUTPUT), CL/V)]
        """
        if isinstance(compartment, str):
            compartment = self.find_compartment(compartment)
        flows = []
        for node in self._g.successors(compartment):
            flow = self.get_flow(compartment, node)
            flows.append((node, flow))
        return flows

    def get_compartment_inflows(self, compartment):
        """Get list of all flows going in to a compartment

        Parameters
        ----------
        compartment : Compartment or str
            Get inflows to this compartment

        Returns
        -------
        list
            Pairs of compartments and symbolic rates

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.get_compartment_inflows("OUTPUT")
        [(Compartment(CENTRAL, dose=Bolus(AMT)), CL/V)]
        """
        if isinstance(compartment, str):
            compartment = self.find_compartment(compartment)
        flows = []
        for node in self._g.predecessors(compartment):
            flow = self.get_flow(node, compartment)
            flows.append((node, flow))
        return flows

    def find_compartment(self, name):
        """Find a compartment using its name

        Parameters
        ----------
        name : str
            Name of compartment to find

        Returns
        -------
        Compartment
            Compartment named name or None if not found

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> central = model.statements.ode_system.find_compartment("CENTRAL")
        >>> central
        Compartment(CENTRAL, dose=Bolus(AMT))
        """
        for comp in self._g.nodes:
            if comp.name == name:
                return comp
        else:
            return None

    def get_n_connected(self, comp):
        """Get the number of compartments connected to a compartment

        Parameters
        ----------
        comp : Compartment or str
            The compartment

        Returns
        -------
        int
            Number of compartments connected to comp

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.get_n_connected("CENTRAL")
        1
        """
        out_comps = {c for c, _ in self.get_compartment_outflows(comp)}
        in_comps = {c for c, _ in self.get_compartment_inflows(comp)}
        return len(out_comps | in_comps)

    @property
    def output_compartment(self):
        """Get the output compartment

        An output compartment is defined to be a compartment that does not have any outward
        flow. A model has to have one and only one output compartment.

        Returns
        -------
        Compartment
            Output compartment of compartmental system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.output_compartment
        Compartment(OUTPUT)
        """
        zeroout = [node for node, out_degree in self._g.out_degree() if out_degree == 0]
        if len(zeroout) == 1:
            return zeroout[0]
        else:
            raise ValueError('More than one or zero output compartments')

    @property
    def dosing_compartment(self):
        """The dosing compartment

        A dosing compartment is a compartment that receives an input dose. Only one dose
        compartment is supported.

        Returns
        -------
        Compartment
            Dosing compartment

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.dosing_compartment
        Compartment(CENTRAL, dose=Bolus(AMT))
        """
        for node in self._g.nodes:
            if node.dose is not None:
                return node
        raise ValueError('No dosing compartment exists')

    @property
    def central_compartment(self):
        """The central compartment

        The central compartment is defined to be the compartment that has an outward flow
        to the output compartment. Only one central compartment is supported.

        Returns
        -------
        Compartment
            Central compartment

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.central_compartment
        Compartment(CENTRAL, dose=Bolus(AMT))
        """
        output = self.output_compartment
        central = next(self._g.predecessors(output))
        return central

    @property
    def peripheral_compartments(self):
        """Find perihperal compartments

        A peripheral compartment is defined as having one flow to the central compartment and
        one flow from the central compartment.

        Returns
        -------
        list of compartments
            Peripheral compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.peripheral_compartments
        []
        """
        central = self.central_compartment
        oneout = {node for node, out_degree in self._g.out_degree() if out_degree == 1}
        onein = {node for node, in_degree in self._g.in_degree() if in_degree == 1}
        cout = {comp for comp in oneout if self.get_flow(comp, central) is not None}
        cin = {comp for comp in onein if self.get_flow(central, comp) is not None}
        peripherals = list(cout & cin)
        # Return in deterministic order
        peripherals = sorted(peripherals, key=lambda comp: comp.name)
        return peripherals

    def find_transit_compartments(self, statements):
        """Find all transit compartments

        Transit compartments are a chain of compartments with the same out rate starting from
        the dose compartment. Because one single transit compartment cannot be distinguished
        from one depot compartment such compartment will be defined to be a depot and not
        a transit compartment.

        Returns
        -------
        list of compartments
            Transit compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.find_transit_compartments(model.statements)
        []
        """
        transits = []
        comp = self.dosing_compartment
        if len(self.get_compartment_inflows(comp)) != 0:
            return transits
        outflows = self.get_compartment_outflows(comp)
        if len(outflows) != 1:
            return transits
        transits.append(comp)
        comp, rate = outflows[0]
        rate = statements.before_odes.full_expression(rate)
        while True:
            if len(self.get_compartment_inflows(comp)) != 1:
                break
            outflows = self.get_compartment_outflows(comp)
            if len(outflows) != 1:
                break
            next_comp, next_rate = outflows[0]
            next_rate = statements.before_odes.full_expression(next_rate)
            if rate != next_rate:
                break
            transits.append(comp)
            comp = next_comp
        # Special case of one transit directly into central is not defined as a transit
        # Also not central itself
        central = self.central_compartment
        if len(transits) == 1 and (
            self.get_flow(transits[0], central) is not None or transits[0] == central
        ):
            return []
        else:
            return transits

    def find_depot(self, statements):
        """Find the depot compartment

        The depot compartment is defined to be the compartment that only has out flow to the
        central compartment, but no flow from the central compartment.

        Returns
        -------
        Compartment
            Depot compartment

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model, set_first_order_absorption
        >>> model = load_example_model("pheno")
        >>> set_first_order_absorption(model)       # doctest: +ELLIPSIS
        <...>
        >>> model.statements.ode_system.find_depot(model.statements)
        Compartment(DEPOT, dose=Bolus(AMT))
        """
        transits = self.find_transit_compartments(statements)
        depot = self._find_depot()
        if depot in transits:
            depot = None
        return depot

    def _find_depot(self):
        central = self.central_compartment
        depot = None
        for to_central, _ in self.get_compartment_inflows(central):
            outflows = self.get_compartment_outflows(to_central)
            if len(outflows) == 1:
                inflows = self.get_compartment_inflows(to_central)
                for in_comp, _ in inflows:
                    if in_comp == central:
                        break
                else:
                    depot = to_central
                    break
        return depot

    @property
    def compartmental_matrix(self):
        """Compartmental matrix of the compartmental system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model, set_first_order_absorption
        >>> import sympy
        >>> model = load_example_model("pheno")
        >>> sympy.pprint(model.statements.ode_system.compartmental_matrix)
        ⎡-CL    ⎤
        ⎢────  0⎥
        ⎢ V     ⎥
        ⎢       ⎥
        ⎢ CL    ⎥
        ⎢ ──   0⎥
        ⎣ V     ⎦
        """
        dod = nx.to_dict_of_dicts(self._g)
        size = len(self._g.nodes)
        f = sympy.zeros(size)
        nodes = list(self._g.nodes)
        output = self.output_compartment
        # Put output last
        nodes.remove(output)
        nodes.append(output)
        for i in range(0, size):
            from_comp = nodes[i]
            diagsum = 0
            for j in range(0, size):
                to_comp = nodes[j]
                try:
                    rate = dod[from_comp][to_comp]['rate']
                except KeyError:
                    rate = 0
                f[j, i] = rate
                diagsum += f[j, i]
            f[i, i] -= diagsum
        return f

    @property
    def amounts(self):
        """Column vector of amounts for all compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model, set_first_order_absorption
        >>> import sympy
        >>> model = load_example_model("pheno")
        >>> sympy.pprint(model.statements.ode_system.amounts)
        ⎡A_CENTRAL⎤
        ⎢         ⎥
        ⎣A_OUTPUT ⎦
        """
        amts = [node.amount for node in self._g.nodes]
        return sympy.Matrix(amts)

    @property
    def names(self):
        """List of the names of all compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model, set_first_order_absorption
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.names
        ['CENTRAL', 'OUTPUT']
        """
        return [node.name for node in self._g.nodes]

    @property
    def zero_order_inputs(self):
        """Vector of all zero order inputs to each compartment

        Example
        -------
        >>> from pharmpy.modeling import load_example_model, set_zero_order_absorption
        >>> import sympy
        >>> model = load_example_model("pheno")
        >>> sympy.pprint(model.statements.ode_system.zero_order_inputs)
        ⎡0⎤
        ⎢ ⎥
        ⎣0⎦
        >>> set_zero_order_absorption(model)    # doctest: +ELLIPSIS
        <...>
        >>> sympy.pprint(model.statements.ode_system.zero_order_inputs)
        ⎡⎧ AMT                ⎤
        ⎢⎪─────  for t < 2⋅MAT⎥
        ⎢⎨2⋅MAT               ⎥
        ⎢⎪                    ⎥
        ⎢⎩  0      otherwise  ⎥
        ⎢                     ⎥
        ⎣          0          ⎦
        """
        inputs = []
        for node in self._g.nodes:
            if node.dose is not None and isinstance(node.dose, Infusion):
                if node.dose.rate is not None:
                    expr = node.dose.rate
                    cond = node.dose.amount / node.dose.rate
                else:
                    expr = node.dose.amount / node.dose.duration
                    cond = node.dose.duration
                infusion_func = sympy.Piecewise((expr, self.t < cond), (0, True))
                inputs.append(infusion_func)
            else:
                inputs.append(0)
        return sympy.Matrix(inputs)

    def to_explicit_system(self, skip_output=False):
        """Get the compartmental system as an explicit ODE system

        Parameters
        ----------
        skip_output : boolean
            Set to true to leave the output compartment out

        Results
        -------
        ExplicitODESystem
            The same ODE system with explicit equations and initial conditions

        Example
        -------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> odes = model.statements.ode_system.to_explicit_system()
        >>> odes
        ⎧d                  -CL⋅A_CENTRAL(t)
        ⎪──(A_CENTRAL(t)) = ─────────────────
        ⎪dt                         V
        ⎨d                 CL⋅A_CENTRAL(t)
        ⎪──(A_OUTPUT(t)) = ───────────────
        ⎪dt                       V
        ⎪A_CENTRAL(0) = AMT
        ⎩A_OUTPUT(0) = 0
        """
        amount_funcs = sympy.Matrix([sympy.Function(amt.name)(self.t) for amt in self.amounts])
        derivatives = sympy.Matrix([sympy.Derivative(fn, self.t) for fn in amount_funcs])
        inputs = self.zero_order_inputs
        a = self.compartmental_matrix @ amount_funcs + inputs
        eqs = [sympy.Eq(lhs, rhs) for lhs, rhs in zip(derivatives, a)]
        ics = {}
        output = self.output_compartment
        for node in self._g.nodes:
            if skip_output and node == output:
                continue
            if node.dose is not None and isinstance(node.dose, Bolus):
                if node.lag_time:
                    time = node.lag_time
                else:
                    time = 0
                ics[sympy.Function(node.amount.name)(time)] = node.dose.amount
            else:
                ics[sympy.Function(node.amount.name)(0)] = sympy.Integer(0)
        if skip_output:
            eqs = eqs[:-1]
        return ExplicitODESystem(eqs, ics)

    def __len__(self):
        """The number of compartments including output"""
        return len(self._g.nodes)

    def _repr_html_(self):
        # Use Unicode art for now. There should be ways of drawing networkx
        s = str(self)
        return f'<pre>{s}</pre>'

    def __repr__(self):
        output = self.output_compartment
        output_box = unicode.Box(output.name)
        central = self.central_compartment
        central_box = unicode.Box(central.name)
        depot = self._find_depot()
        current = self.dosing_compartment
        if depot:
            comp = depot
        else:
            comp = central
        transits = []
        while current != comp:
            transits.append(current)
            current = self.get_compartment_outflows(current)[0][0]
        periphs = self.peripheral_compartments
        nrows = 1 + 2 * len(periphs)
        ncols = 2 * len(transits) + (2 if depot else 0) + 3
        grid = unicode.Grid(nrows, ncols)
        if nrows == 1:
            main_row = 0
        else:
            main_row = 2
        col = 0
        for transit in transits:
            grid.set(main_row, col, unicode.Box(transit.name))
            col += 1
            grid.set(
                main_row, col, unicode.Arrow(str(self.get_compartment_outflows(transit)[0][1]))
            )
            col += 1
        if depot:
            grid.set(main_row, col, unicode.Box(depot.name))
            col += 1
            grid.set(main_row, col, unicode.Arrow(str(self.get_compartment_outflows(depot)[0][1])))
            col += 1
        central_col = col
        grid.set(main_row, col, central_box)
        col += 1
        grid.set(main_row, col, unicode.Arrow(str(self.get_flow(central, output))))
        col += 1
        grid.set(main_row, col, output_box)
        if periphs:
            grid.set(0, central_col, unicode.Box(periphs[0].name))
            grid.set(
                1,
                central_col,
                unicode.DualVerticalArrows(
                    str(self.get_flow(central, periphs[0])), str(self.get_flow(periphs[0], central))
                ),
            )
        if len(periphs) > 1:
            grid.set(4, central_col, unicode.Box(periphs[1].name))
            grid.set(
                3,
                central_col,
                unicode.DualVerticalArrows(
                    str(self.get_flow(periphs[1], central)), str(self.get_flow(central, periphs[1]))
                ),
            )

        dose = self.dosing_compartment.dose
        s = str(dose) + '\n' + str(grid).rstrip()
        return s


class Compartment:
    """Compartment for a compartmental system

    Attributes
    ----------
    name : str
        Compartment name
    dose : Dose
        Dose object for dose into this compartment. Default None for no dose.
    lag_time : Expression
        Lag time for doses entering this compartment. Default 0

    Examples
    --------
    >>> from pharmpy import Bolus, Compartment
    >>> comp = Compartment("CENTRAL")
    >>> comp
    Compartment(CENTRAL)
    >>> comp = Compartment("DEPOT", lag_time="ALAG")
    >>> comp
    Compartment(DEPOT, lag_time=ALAG)
    >>> dose = Bolus("AMT")
    >>> comp = Compartment("DEPOT", dose=dose)
    >>> comp
    Compartment(DEPOT, dose=Bolus(AMT))
    """

    def __init__(self, name, dose=None, lag_time=0):
        self.name = name
        self.dose = dose
        self.lag_time = lag_time

    @property
    def amount(self):
        """Symbol for the amount in the compartment

        Examples
        --------
        >>> from pharmpy import Compartment
        >>> comp = Compartment("CENTRAL")
        >>> comp.amount
        A_CENTRAL
        """
        return symbols.symbol(f'A_{self.name}')

    @property
    def lag_time(self):
        return self._lag_time

    @lag_time.setter
    def lag_time(self, value):
        self._lag_time = sympy.sympify(value)

    @property
    def free_symbols(self):
        """Get set of all free symbols in the compartment

        Examples
        --------
        >>> from pharmpy import Bolus, Compartment
        >>> dose = Bolus("AMT")
        >>> comp = Compartment("CENTRAL", dose=dose, lag_time="ALAG")
        >>> comp.free_symbols  # doctest: +SKIP
        {ALAG, AMT}
        """
        symbs = set()
        if self.dose is not None:
            symbs |= self.dose.free_symbols
        symbs |= self.lag_time.free_symbols
        return symbs

    def subs(self, substitutions):
        """Substitute expressions or symbols in compartment

        Examples
        --------
        >>> from pharmpy import Bolus, Compartment
        >>> dose = Bolus("AMT")
        >>> comp = Compartment("CENTRAL", dose=dose)
        >>> comp.subs({"AMT": "DOSE"})
        >>> comp
        Compartment(CENTRAL, dose=Bolus(DOSE))
        """
        if self.dose is not None:
            self.dose.subs(substitutions)
        self.lag_time.subs(substitutions)

    def __eq__(self, other):
        return (
            isinstance(other, Compartment)
            and self.name == other.name
            and self.dose == other.dose
            and self.lag_time == other.lag_time
        )

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        lag = '' if self.lag_time == 0 else f', lag_time={self.lag_time}'
        dose = '' if self.dose is None else f', dose={self.dose}'
        return f'Compartment({self.name}{dose}{lag})'


class Bolus:
    """A Bolus dose

    Attributes
    ----------
    amount : symbol
        Symbolic amount of dose

    Examples
    --------
    >>> from pharmpy import Bolus
    >>> dose = Bolus("AMT")
    >>> dose
    Bolus(AMT)
    """

    def __init__(self, amount):
        self.amount = sympy.sympify(amount)

    @property
    def free_symbols(self):
        """Get set of all free symbols in the dose

        Examples
        --------
        >>> from pharmpy import Bolus
        >>> dose = Bolus("AMT")
        >>> dose.free_symbols
        {AMT}
        """
        return {self.amount}

    def subs(self, substitutions):
        """Substitute expressions or symbols in dose

        Parameters
        ----------
        substitutions : dict
            Dictionary of from, to pairs

        Examples
        --------
        >>> from pharmpy import Bolus
        >>> dose = Bolus("AMT")
        >>> dose.subs({'AMT': 'DOSE'})
        >>> dose
        Bolus(DOSE)
        """
        self.amount = self.amount.subs(substitutions, simultaneous=True)

    def __deepcopy__(self, memo):
        newone = type(self)(self.amount)
        return newone

    def __eq__(self, other):
        return isinstance(other, Bolus) and self.amount == other.amount

    def __repr__(self):
        return f'Bolus({self.amount})'


class Infusion:
    """An infusion dose

    Attributes
    ----------
    amount : expression
        Symbolic amount of dose
    rate : expression
        Symbolic rate. Mutually excluseive with duration
    duration : expression
        Symbolic duration. Mutually excluseive with rate

    Examples
    --------
    >>> from pharmpy import Infusion
    >>> dose = Infusion("AMT", duration="D1")
    >>> dose
    Infusion(AMT, duration=D1)
    >>> dose = Infusion("AMT", rate="R1")
    >>> dose
    Infusion(AMT, rate=R1)
    """

    def __init__(self, amount, rate=None, duration=None):
        if rate is None and duration is None:
            raise ValueError('Need rate or duration for Infusion')
        self.rate = sympy.sympify(rate)
        self.duration = sympy.sympify(duration)
        self.amount = sympy.sympify(amount)

    @property
    def free_symbols(self):
        """Get set of all free symbols in the dose

        Examples
        --------
        >>> from pharmpy import Infusion
        >>> dose = Infusion("AMT", rate="RATE")
        >>> dose.free_symbols   # doctest: +SKIP
        {AMT, RATE}
        """
        if self.rate is not None:
            symbs = self.rate.free_symbols
        else:
            symbs = self.duration.free_symbols
        return symbs | self.amount.free_symbols

    def subs(self, substitutions):
        """Substitute expressions or symbols in dose

        Parameters
        ----------
        substitutions : dict
            Dictionary of from, to pairs

        Examples
        --------
        >>> from pharmpy import Infusion
        >>> dose = Infusion("AMT", duration="DUR")
        >>> dose.subs({'DUR': 'D1'})
        >>> dose
        Infusion(AMT, duration=D1)
        """
        self.amount = self.amount.subs(substitutions, simultaneous=True)
        if self.rate is not None:
            self.rate = self.rate.subs(substitutions, simultaneous=True)
        else:
            self.duration = self.duration.subs(substitutions, simultaneous=True)

    def __deepcopy__(self, memo):
        new = type(self)(self.amount, rate=self.rate, duration=self.duration)
        return new

    def __eq__(self, other):
        return (
            isinstance(other, Infusion)
            and self.rate == other.rate
            and self.duration == other.duration
            and self.amount == other.amount
        )

    def __repr__(self):
        if self.rate is not None:
            arg = f'rate={self.rate}'
        else:
            arg = f'duration={self.duration}'
        return f'Infusion({self.amount}, {arg})'


class ModelStatements(MutableSequence):
    """A sequence of symbolic statements describing the model

    Two types of statements are supported: Assignment and ODESystem.
    A ModelStatements object can have 0 or 1 ODESystem. The order of
    the statements is significant and the same symbol can be assigned
    to multiple times.
    """

    def __init__(self, statements=None):
        if isinstance(statements, ModelStatements):
            self._statements = copy.deepcopy(statements._statements)
        elif statements is None:
            self._statements = []
        else:
            self._statements = list(statements)

    def __getitem__(self, ind):
        return self._statements[ind]

    def __setitem__(self, ind, value):
        self._statements[ind] = value

    def __delitem__(self, ind):
        del self._statements[ind]

    def __len__(self):
        return len(self._statements)

    def insert(self, ind, value):
        self._statements.insert(ind, value)

    @property
    def free_symbols(self):
        """Get a set of all free symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.free_symbols   # doctest: +SKIP
        {AMT, APGR, A_CENTRAL, BTIME, CL, DV, EPS(1), ETA(1), ETA(2), F, IPRED, IRES, IWRES, S1,
        TAD, THETA(1), THETA(2), THETA(3), TIME, TVCL, TVV, V, W, WGT, Y, t}

        """
        symbols = set()
        for assignment in self:
            symbols |= assignment.free_symbols
        return symbols

    @property
    def ode_system(self):
        """Returns the ODE system of the model or None if the model doesn't have an ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system
        Bolus(AMT)
        ┌───────┐       ┌──────┐
        │CENTRAL│──CL/V→│OUTPUT│
        └───────┘       └──────┘
        """
        for s in self:
            if isinstance(s, ODESystem):
                return s
        return None

    @property
    def before_odes(self):
        """All statements before the ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.before_odes
                 ⎧TIME  for AMT > 0
                 ⎨
        BTIME := ⎩ 0     otherwise
        TAD := -BTIME + TIME
        TVCL := THETA(1)⋅WGT
        TVV := THETA(2)⋅WGT
               ⎧TVV⋅(THETA(3) + 1)  for APGR < 5
               ⎨
        TVV := ⎩       TVV           otherwise
                    ETA(1)
        CL := TVCL⋅ℯ
                  ETA(2)
        V := TVV⋅ℯ
        S₁ := V
        """
        sset = ModelStatements()
        for s in self:
            if isinstance(s, ODESystem):
                break
            sset.append(s)
        return sset

    @property
    def after_odes(self):
        """All statements after the ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.after_odes
             A_CENTRAL
             ─────────
        F :=     S₁
        W := F
        Y := EPS(1)⋅W + F
        IPRED := F
        IRES := DV - IPRED
                  IRES
                  ────
        IWRES :=   W
        """
        sset = ModelStatements()
        found = False
        if self.ode_system is None:
            return self
        for s in self:
            if isinstance(s, ODESystem):
                found = True
            elif found:
                sset.append(s)
        return sset

    def subs(self, substitutions):
        """Substitute symbols in all statements.

        Parameters
        ----------
        substitutions : dict
            Old-new pairs(can be type str or sympy symbol)

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.subs({'WGT': 'WT'})

        """
        for statement in self:
            statement.subs(substitutions)

    def find_assignment(self, symbol):
        """Returns last assignment of symbol

        Parameters
        ----------
        symbol : Symbol or str
            Symbol to look for

        Returns
        -------
        Assignment
            An Assignment or None if no assignment to symbol exists

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.find_assignment("CL")
                    ETA(1)
        CL := TVCL⋅ℯ
        """
        symbol = sympy.sympify(symbol)
        assignment = None
        for statement in self:
            if isinstance(statement, Assignment):
                if statement.symbol == symbol:
                    assignment = statement
        return assignment

    def reassign(self, symbol, expression):
        """Reassign symbol to expression

        Set symbol to be expression and remove all previous assignments of symbol

        Parameters
        ----------
        symbol : Symbol or str
            Symbol to reassign
        expression : Expression or str
            The new expression to assign to symbol

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.reassign("CL", "TVCL + eta")
        """
        if isinstance(symbol, str):
            symbol = sympy.sympify(symbol)
        if isinstance(expression, str):
            expression = sympy.sympify(expression)

        last = True
        for i, stat in zip(range(len(self) - 1, -1, -1), reversed(self)):
            if isinstance(stat, Assignment) and stat.symbol == symbol:
                if last:
                    stat.expression = expression
                    last = False
                else:
                    del self[i]

    def _create_dependency_graph(self):
        """Create a graph of dependencies between statements"""
        graph = nx.DiGraph()
        for i in range(len(self) - 1, -1, -1):
            rhs = self[i].rhs_symbols
            for s in rhs:
                for j in range(i - 1, -1, -1):
                    if (
                        isinstance(self[j], Assignment)
                        and self[j].symbol == s
                        or isinstance(self[j], ODESystem)
                        and s in self[j].amounts
                    ):
                        graph.add_edge(i, j)
        return graph

    def dependencies(self, symbol):
        """Find all dependencies of a symbol

        Parameters
        ----------
        symbol : Symbol or str
            Input symbol

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.dependencies("CL")   # doctest: +SKIP
        {ETA(1), THETA(1), WGT}
        """
        symbol = sympy.sympify(symbol)
        for i in range(len(self) - 1, -1, -1):
            if (
                isinstance(self[i], Assignment)
                and self[i].symbol == symbol
                or isinstance(self[i], ODESystem)
                and symbol in self[i].amounts
            ):
                break
        else:
            raise KeyError(f"Could not find symbol {symbol}")
        g = self._create_dependency_graph()
        symbs = self[i].rhs_symbols
        if i == 0:
            # Special case for models with only one statement
            return symbs
        for j, _ in nx.bfs_predecessors(g, i, sort_neighbors=lambda x: reversed(sorted(x))):
            if isinstance(self[j], Assignment):
                symbs -= {self[j].symbol}
            elif isinstance(self[j], ODESystem):
                symbs -= set(self[j].amounts)
            symbs |= self[j].rhs_symbols
        return symbs

    def remove_symbol_definitions(self, symbols, statement):
        """Remove symbols and dependencies not used elsewhere

        If the statement no longer depends on the specified
        symbols, this method will make sure that the definitions
        of these symbols will be removed unless they are dependencies
        of other statements.

        Parameters
        ----------
        symbols : iterable
            Iterable of symbols no longer used in the statement
        statement : Statement
            Statement from which the symbols were removed
        """
        graph = self._create_dependency_graph()
        removed_ind = self.index(statement)
        # Statements defining symbols and dependencies
        candidates = set()
        for s in symbols:
            for i in range(removed_ind - 1, -1, -1):
                stat = self[i]
                if isinstance(stat, Assignment) and stat.symbol == s:
                    candidates.add(i)
                    break
        for i in candidates.copy():
            if i in graph:
                candidates |= set(nx.dfs_preorder_nodes(graph, i))
        # All statements needed for removed_ind
        if removed_ind in graph:
            keep = {down for _, down in nx.dfs_edges(graph, removed_ind)}
        else:
            keep = set()
        candidates -= keep
        # Other dependencies after removed_ind
        additional = {down for up, down in graph.edges if up > removed_ind and down in candidates}
        for add in additional.copy():
            if add in graph:
                additional |= set(nx.dfs_preorder_nodes(graph, add))
        remove = candidates - additional
        for i in reversed(sorted(remove)):
            del self[i]

    def full_expression(self, expression):
        """Expand an expression into its full definition

        Parameters
        ----------
        expression : expression or str
            Expression to expand. A string will be converted to an expression.

        Return
        ------
        expression
            Expanded expression

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.before_odes.full_expression("CL")
        THETA(1)*WGT*exp(ETA(1))
        """
        if isinstance(expression, str):
            expression = sympy.sympify(expression)
        for statement in reversed(self):
            if isinstance(statement, ODESystem):
                raise ValueError(
                    "ODESystem not supported by full_expression. Use the properties before_odes "
                    "or after_odes."
                )
            expression = expression.subs({statement.symbol: statement.expression})
        return expression

    def insert_before_odes(self, statement):
        """Insert a statement just before the ODE system or at the end of the model

        Parameters
        ----------
        statement : Statement
            Statement to insert

        Examples
        --------
        >>> from pharmpy import Assignment
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> a = Assignment("WGT_G", "WGT*1000")
        >>> model.statements.insert_before_odes(a)
        """
        for i, s in enumerate(self):
            if isinstance(s, ODESystem):
                break
        else:
            i += 1
        self.insert(i, statement)

    def to_compartmental_system(self):
        """Convert ODE system to a compartmental system

        raise if not possible
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.to_explicit_system()
        >>> model.statements.to_compartmental_system()
        >>> model.statements.ode_system
        Bolus(AMT)
        ┌───────┐       ┌──────┐
        │CENTRAL│──CL/V→│OUTPUT│
        └───────┘       └──────┘
        """
        for i, s in enumerate(self):
            if isinstance(s, ExplicitODESystem):
                new = s.to_compartmental_system()
                self[i] = new
                return

    def to_explicit_system(self):
        """Convert ODE system to an explicit ODE system

        Example
        -------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system
        Bolus(AMT)
        ┌───────┐       ┌──────┐
        │CENTRAL│──CL/V→│OUTPUT│
        └───────┘       └──────┘
        >>> model.statements.to_explicit_system()
        >>> model.statements.ode_system
        ⎧d                  -CL⋅A_CENTRAL(t)
        ⎪──(A_CENTRAL(t)) = ─────────────────
        ⎪dt                         V
        ⎨d                 CL⋅A_CENTRAL(t)
        ⎪──(A_OUTPUT(t)) = ───────────────
        ⎪dt                       V
        ⎪A_CENTRAL(0) = AMT
        ⎩A_OUTPUT(0) = 0
        """
        for i, s in enumerate(self):
            if isinstance(s, CompartmentalSystem):
                new = s.to_explicit_system()
                self[i] = new
                return

    def copy(self):
        """Create a copy of the ModelStatements object"""
        return copy.deepcopy(self)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        else:
            for first, second in zip(self, other):
                if first != second:
                    return False
        return True

    def __repr__(self):
        return '\n'.join([repr(statement) for statement in self])

    def _repr_html_(self):
        html = r'\begin{align*}'
        for statement in self:
            if hasattr(statement, '_repr_html_'):
                html += '\\end{align*}'
                s = statement._repr_html_()
                html += s + '\\begin{align*}'
            else:
                s = f'${statement._repr_latex_()}$'
                s = s.replace(':=', '&:=')
                s = s.replace('$', '')
                s = s + r'\\'
                html += s
        return html + '\\end{align*}'



# -*- encoding: utf-8 -*-

from . import record


class RawRecord(record.Record):
    """A record that just keeps contents unparsed.
    Used for unknown records and for anything coming before the first record
    """

    def __init__(self, content):
        self.content = content
        self.raw_name = ''

    def __str__(self):
        return self.raw_name + self.content



from pharmpy.plugins.nonmem.nmtran_parser import NMTranParser


def test_simple_parse():
    parser = NMTranParser()

    model = parser.parse('$PROBLEM MYPROB\n')

    assert len(model.records) == 1
    assert type(model.records[0]).__name__ == 'ProblemRecord'
    assert str(model) == '$PROBLEM MYPROB\n'

    model2_str = ';Comment\n   $PROBLEM     TW2\n'
    model2 = parser.parse(model2_str)

    assert len(model2.records) == 2
    assert type(model2.records[0]).__name__ == 'RawRecord'
    assert type(model2.records[1]).__name__ == 'ProblemRecord'

    assert str(model2) == model2_str


def test_round_trip(pheno_path):
    parser = NMTranParser()

    with open(pheno_path, 'r') as fh:
        content = fh.read()
    model = parser.parse(content)
    assert str(model) == content



import pharmpy.tools
from pharmpy.model import Model
from pharmpy.plugins.nonmem.run import execute_model
from pharmpy.workflows import Task, Workflow


def create_workflow(models=None, n=None):
    """
    If models is None and n is None: fit one unknown model
    If models is model object (n must be None): fit that one model
    If models is list of models (n must be None): fit these models
    If models is None and n is an int: fit n unknown models
    """
    wf = Workflow()
    wf.name = "modelfit"
    if models is None:
        if n is None:
            task = Task('run', execute_model)
            wf.add_task(task)
        else:
            for i in range(n):
                task = Task(f'run{i}', execute_model)
                wf.add_task(task)
    elif isinstance(models, Model):
        task = Task('run', execute_model, models)
    else:
        for i, model in enumerate(models):
            task = Task(f'run{i}', execute_model, model)
            wf.add_task(task)
    task_result = Task('results', post_process_results)
    wf.add_task(task_result, predecessors=wf.output_tasks)
    return wf


# FIXME: These two aren't really needed
def create_single_fit_workflow(model=None):
    wf = Workflow()
    if model is None:
        task = Task('run', execute_model)
    else:
        task = Task('run', execute_model, model)
    wf.add_task(task)
    return wf


def create_multiple_fit_workflow(models=None, n=None):
    """Either specify models or n"""
    wf = Workflow()
    if models is None:
        for i in range(n):
            task = Task(f'run{i}', execute_model)
            wf.add_task(task)
    else:
        for i, model in enumerate(models):
            task = Task(f'run{i}', execute_model, model)
            wf.add_task(task)
    return wf


class Modelfit(pharmpy.tools.Tool):
    def __init__(self, models, **kwargs):
        self.models = models
        super().__init__(**kwargs)

    def run(self):
        wf_fit = create_multiple_fit_workflow(self.models)
        task_result = Task('results', post_process_results)
        wf_fit.add_task(task_result, predecessors=wf_fit.output_tasks)
        for model in self.models:
            model.dataset
            model.database = self.database.model_database
        fit_models = self.dispatcher.run(wf_fit, self.database)
        for i in range(len(fit_models)):
            self.models[i].modelfit_results = fit_models[i].modelfit_results


def post_process_results(*models):
    return models



import pandas as pd

from pharmpy.results import Results


class AggregatedModelfitResults(Results):
    def __init__(self, ofv=None, parameter_estimates=None):
        self.ofv = ofv
        self.parameter_estimates = parameter_estimates


def calculate_results(models):
    names = [model.name for model in models]

    ofvs = [model.modelfit_results.ofv for model in models]
    ofv = pd.Series(ofvs, index=names)
    ofv.name = 'OFV'

    params = pd.DataFrame(
        [model.modelfit_results.parameter_estimates for model in models], index=names
    )

    res = AggregatedModelfitResults(ofv=ofv, parameter_estimates=params)
    return res



import pytest
from sympy import exp

from pharmpy import Model
from pharmpy.modeling.covariate_effect import CovariateEffect, _choose_param_inits
from pharmpy.symbols import symbol as S


@pytest.mark.parametrize(
    'cov_eff,symbol,expression',
    [
        (
            CovariateEffect.exponential(),
            S('CLWGT'),
            exp(S('COVEFF1') * (S('WGT') - S('WGT_MEDIAN'))),
        ),
        (CovariateEffect.power(), S('CLWGT'), (S('WGT') / S('WGT_MEDIAN')) ** S('COVEFF1')),
        (CovariateEffect.linear(), S('CLWGT'), 1 + S('COVEFF1') * (S('WGT') - S('WGT_MEDIAN'))),
    ],
)
def test_apply(cov_eff, symbol, expression):
    cov_eff.apply(
        parameter='CL',
        covariate='WGT',
        thetas={'theta': 'COVEFF1'},
        statistics={'mean': 1, 'median': 1, 'std': 1},
    )

    assert cov_eff.template.symbol == symbol
    assert cov_eff.template.expression == expression


@pytest.mark.parametrize(
    'cov_eff, init, lower, upper', [('exp', 0.001, -0.8696, 0.8696), ('pow', 0.001, -100, 100000)]
)
def test_choose_param_inits(pheno_path, cov_eff, init, lower, upper):
    model = Model(pheno_path)

    inits = _choose_param_inits(cov_eff, model.dataset, 'WGT')

    assert inits['init'] == init
    assert inits['lower'] == lower
    assert inits['upper'] == upper



class ToolDatabase:
    """Database of results, metadata and run files for one run of a tool

    This corresponds to the run directory of PsN

     Attribute: model_database
    """

    def __init__(self, toolname):
        self.toolname = toolname


class ModelDatabase:
    """Database of results for particular model runs

    This corresponds to the m1 directory and copying up files in PsN
    """

    def store_local_file(self, model, path):
        """Store a file from the local machine"""
        raise NotImplementedError()



from itertools import combinations

from pharmpy.modeling import copy_model
from pharmpy.modeling.block_rvs import create_joint_distribution
from pharmpy.tools.modelfit import create_multiple_fit_workflow
from pharmpy.workflows import Task, Workflow


def brute_force(model):
    wf = Workflow()
    eta_combos = _get_iiv_combinations(model)

    for i, combo in enumerate(eta_combos, 1):
        model_name = f'iiv_candidate{i}'
        task_copy = Task('copy', copy, model_name)
        wf.add_task(task_copy)

        task_joint_dist = Task('create_joint_dist', create_joint_dist, combo)
        wf.add_task(task_joint_dist, predecessors=task_copy)

    wf_fit = create_multiple_fit_workflow(n=len(eta_combos))
    wf.insert_workflow(wf_fit)
    return wf


def _create_joint_distribution(rvs, model):
    return create_joint_distribution(model, rvs)


def _get_iiv_combinations(model):
    eta_names = model.random_variables.etas.names
    no_of_etas = len(eta_names)
    eta_combos = []
    for i in range(2, no_of_etas + 1):
        eta_combos += [list(combo) for combo in combinations(eta_names, i)]
    return eta_combos


def create_joint_dist(list_of_etas, model):
    return create_joint_distribution(model, list_of_etas)


def copy(name, model):
    model_copy = copy_model(model, name)
    return model_copy



from io import StringIO

import numpy as np
import pandas as pd
import pytest

from pharmpy import Model
from pharmpy.results import read_results
from pharmpy.tools.qa.results import calculate_results, psn_qa_results
from pharmpy.tools.resmod.results import psn_resmod_results


def test_add_etas(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    add_etas = Model(testdata / 'nonmem' / 'qa' / 'add_etas_linbase.mod')
    res = calculate_results(orig, base, add_etas_model=add_etas, etas_added_to=['CL', 'V'])
    correct = """added,new_sd,orig_sd
ETA(1),True,0.338974,0.333246
ETA(2),True,0.449430,0.448917
CL,False,0.010001,NaN
V,False,0.010000,NaN
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.add_etas_parameters, correct, atol=1e-6)

    assert res.dofv['dofv']['parameter_variability', 'add_etas', np.nan] == pytest.approx(
        730.89472681373070 - 730.84697789365532
    )
    assert res.dofv['df']['parameter_variability', 'add_etas', np.nan] == 2


def test_fullblock(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    fb = Model(testdata / 'nonmem' / 'qa' / 'fullblock.mod')
    res = calculate_results(orig, base, fullblock_model=fb)
    correct = """,new,old
"OMEGA(1,1)",0.486600,0.333246
"OMEGA(2,1)",0.846728,NaN
"OMEGA(2,2)",0.423262,0.448917
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.fullblock_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'fullblock', np.nan] == pytest.approx(
        730.89472681373070 - 706.36113798726512
    )
    assert res.dofv['df']['parameter_variability', 'fullblock', np.nan] == 1

    res = calculate_results(orig, base, fullblock_model=None)
    assert res.fullblock_parameters is None


def test_boxcox(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    bc = Model(testdata / 'nonmem' / 'qa' / 'boxcox.mod')
    res = calculate_results(orig, base, boxcox_model=bc)
    correct = """lambda,new_sd,old_sd
ETA(1),-1.581460,0.296257,0.333246
ETA(2),0.645817,0.429369,0.448917
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.boxcox_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'boxcox', np.nan] == pytest.approx(
        730.89472681373070 - 721.78812733817688
    )
    assert res.dofv['df']['parameter_variability', 'boxcox', np.nan] == 2

    res = calculate_results(orig, base, boxcox_model=None)
    assert res.boxcox_parameters is None


def test_tdist(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    td = Model(testdata / 'nonmem' / 'qa' / 'tdist.mod')
    res = calculate_results(orig, base, tdist_model=td)
    correct = """df,new_sd,old_sd
ETA(1),3.77,0.344951,0.333246
ETA(2),3.77,0.400863,0.448917
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.tdist_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'tdist', np.nan] == pytest.approx(
        730.89472681373070 - 729.45800311609150
    )
    assert res.dofv['df']['parameter_variability', 'tdist', np.nan] == 2

    res = calculate_results(orig, base, tdist_model=None)


def test_iov(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    iov = Model(testdata / 'nonmem' / 'qa' / 'iov.mod')
    res = calculate_results(orig, base, iov_model=iov)
    correct = """new_iiv_sd,orig_iiv_sd,iov_sd
ETA(1),0.259560,0.333246,0.555607
ETA(2),0.071481,0.448917,0.400451
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0])
    pd.testing.assert_frame_equal(res.iov_parameters, correct)

    assert res.dofv['dofv']['parameter_variability', 'iov', np.nan] == pytest.approx(42.314986)
    assert res.dofv['df']['parameter_variability', 'iov', np.nan] == 2


def test_scm(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    scm_res = read_results(testdata / 'nonmem' / 'qa' / 'scm_results.json')
    res = calculate_results(orig, base, scm_results=scm_res)
    correct = """,,dofv,coeff
ETA(1),APGR,2.48792,-0.033334
ETA(1),WGT,0.48218,0.052342
ETA(2),APGR,0.59036,0.008371
ETA(2),WGT,0.00887,-0.003273
"""
    correct = pd.read_csv(StringIO(correct), index_col=[0, 1])
    correct.index.set_names(['parameter', 'covariate'], inplace=True)
    pd.testing.assert_frame_equal(res.covariate_effects, correct, atol=1e-6)
    assert res.dofv['dofv']['covariates', 'ET1APGR-2', np.nan] == pytest.approx(2.48792)
    assert res.dofv['df']['covariates', 'ET1APGR-2', np.nan] == 1


def test_resmod(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    resmod_res = read_results(testdata / 'nonmem' / 'qa' / 'resmod_results.json')
    res = calculate_results(orig, base, resmod_idv_results=resmod_res)
    assert list(res.residual_error['additional_parameters']) == [2, 2, 1, 1, 1, 1]
    assert list(res.residual_error['dOFV']) == [13.91, 8.03, 5.53, 3.34, 1.31, 0.03]
    assert res.residual_error.index.tolist() == [
        (1, 'dtbs'),
        (1, 'time_varying'),
        (1, 'tdist'),
        (1, 'autocorrelation'),
        (1, 'IIV_on_RUV'),
        (1, 'power'),
    ]
    assert res.dofv['dofv']['residual_error_model', 'dtbs', 1] == pytest.approx(13.91)


def test_resmod_dvid(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    resmod_res = psn_resmod_results(testdata / 'psn' / 'resmod_dir2')
    res = calculate_results(orig, base, resmod_idv_results=resmod_res)
    assert res.residual_error.loc[("4", "tdist"), 'dOFV'] == 2.41


def test_psn_qa_results(testdata):
    path = testdata / 'psn' / 'qa_dir1'
    res = psn_qa_results(path)
    correct = """idv,dvid,binid,binmin,binmax,cwres,cpred
TIME,1,1,0.00,2.00,0.28,-6
TIME,1,2,2.00,2.55,0.12,-2
TIME,1,3,2.55,11.00,-0.29,6
TIME,1,4,11.00,47.25,0.04,-1
TIME,1,5,47.25,63.50,-0.39,7
TIME,1,6,63.50,83.10,0.20,-4
TIME,1,7,83.10,112.30,0.11,-2
TIME,1,8,112.30,135.50,-0.29,5
TIME,1,9,135.50,159.80,0.19,-4
TIME,1,10,159.80,390.00,-0.02,1
TAD,1,1,0.00,1.50,0.06,-1
TAD,1,2,1.50,2.00,0.41,-8
TAD,1,3,2.00,3.00,-0.13,3
TAD,1,4,3.00,6.00,-0.06,1
TAD,1,5,6.00,11.00,-0.18,3
TAD,1,6,11.00,11.50,0.54,-10
TAD,1,7,11.50,11.70,-0.24,4
TAD,1,8,11.70,14.00,0.06,-1
TAD,1,9,14.00,258.00,0.07,-1
PRED,1,1,8.00,17.67,0.19,-5
PRED,1,2,17.67,19.50,0.10,-1
PRED,1,3,19.50,20.13,-0.17,3
PRED,1,4,20.13,21.39,-0.01,0
PRED,1,5,21.39,24.32,0.24,-4
PRED,1,6,24.32,26.63,0.06,-1
PRED,1,7,26.63,28.70,0.05,-1
PRED,1,8,28.70,31.28,-0.05,1
PRED,1,9,31.28,36.34,0.07,0
PRED,1,10,36.34,54.00,-0.47,9
"""

    correct = pd.read_csv(StringIO(correct), index_col=[0, 1, 2])
    pd.testing.assert_frame_equal(res.structural_bias, correct, atol=1e-6)


def test_simeval(testdata):
    orig = Model(testdata / 'nonmem' / 'pheno.mod')
    base = Model(testdata / 'nonmem' / 'qa' / 'pheno_linbase.mod')
    simeval_res = read_results(testdata / 'nonmem' / 'qa' / 'simeval_results.json')
    cdd_res = read_results(testdata / 'nonmem' / 'qa' / 'cdd_results.json')
    calculate_results(orig, base, simeval_results=simeval_res, cdd_results=cdd_res)



import copy
from collections.abc import MutableSequence

import pandas as pd
import sympy

import pharmpy.symbols as symbols
from pharmpy.math import is_near_target
from pharmpy.utils import parse_units


class Parameters(MutableSequence):
    """A collection of parameters

    Class representing a group of parameters. Usually all parameters in a model.
    This class give a ways of displaying, summarizing and manipulating
    more than one parameter at a time.

    Specific parameters can be found using indexing on the parameter name

    Example
    -------

    >>> from pharmpy import Parameters, Parameter
    >>> par1 = Parameter("x", 0)
    >>> par2 = Parameter("y", 1)
    >>> pset = Parameters([par1, par2])
    >>> pset["x"]
    Parameter("x", 0, lower=-oo, upper=oo, fix=False)

    >>> "x" in pset
    True

    >>> pset.append(Parameter("z", 0.1))
    >>> len(pset)
    3

    >>> del pset["x"]
    >>> len(pset)
    2
    """

    def __init__(self, params=None):
        if isinstance(params, Parameters):
            self._params = copy.deepcopy(params._params)
        elif params is None:
            self._params = []
        else:
            self._params = list(params)
        names = set()
        for p in self._params:
            if not isinstance(p, Parameter):
                raise ValueError(f'Can not add variable of type {type(p)} to Parameters')
            if p.name in names:
                raise ValueError(
                    f'Parameter names must be unique. Parameter "{p.name}" '
                    'was added more than once to Parameters'
                )
            names.add(p.name)

    def __len__(self):
        return len(self._params)

    def _lookup_param(self, ind, insert=False):
        if isinstance(ind, sympy.Symbol):
            ind = ind.name
        if isinstance(ind, str):
            for i, param in enumerate(self._params):
                if ind == param.name:
                    return i, param
            raise KeyError(f'Could not find {ind} in Parameters')
        elif isinstance(ind, Parameter):
            try:
                i = self._params.index(ind)
            except ValueError:
                raise KeyError(f'Could not find {ind.name} in Parameters')
            return i, ind
        if insert:
            # Must allow for inserting after last element.
            return ind, None
        else:
            return ind, self._params[ind]

    def __getitem__(self, ind):
        if isinstance(ind, list):
            params = []
            for i in ind:
                index, param = self._lookup_param(i)
                params.append(self[index])
            return Parameters(params)
        else:
            _, param = self._lookup_param(ind)
            return param

    def __setitem__(self, ind, value):
        if not isinstance(value, Parameter):
            raise ValueError(f"Can not set variable of type {type(value)} to Parameters.")
        i, rv = self._lookup_param(ind)
        if rv.name != value.name and value.name in self.names:
            raise ValueError(
                f"Cannot set parameter with already existing name {value.name} " "into Parameters."
            )
        self._params[i] = value

    def __delitem__(self, ind):
        i, _ = self._lookup_param(ind)
        del self._params[i]

    def insert(self, ind, value):
        if not isinstance(value, Parameter):
            raise ValueError(
                f'Trying to insert {type(value)} into Parameters. Must be of type Parameter.'
            )
        i, _ = self._lookup_param(ind, insert=True)
        if value.name in self.names:
            raise ValueError(
                f"Cannot insert parameter with alread existing name {value.name} "
                "into Parameters."
            )
        self._params.insert(i, value)

    def __contains__(self, ind):
        try:
            _, _ = self._lookup_param(ind)
        except KeyError:
            return False
        return True

    def to_dataframe(self):
        """Create a dataframe with a summary of all Parameters

        Returns
        -------
        DataFrame
            A dataframe with one row per parameter.
            The columns are value, lower, upper and fix
            Row Index is the names

        Example
        -------
        >>> from pharmpy import Parameters, Parameter
        >>> par1 = Parameter("CL", 1, lower=0, upper=10)
        >>> par2 = Parameter("V", 10, lower=0, upper=100)
        >>> pset = Parameters([par1, par2])
        >>> pset.to_dataframe()
            value  lower  upper    fix
        CL      1      0     10  False
        V      10      0    100  False

        """
        symbols = [param.name for param in self._params]
        values = [param.init for param in self._params]
        lower = [param.lower for param in self._params]
        upper = [param.upper for param in self._params]
        fix = [param.fix for param in self._params]
        return pd.DataFrame(
            {'value': values, 'lower': lower, 'upper': upper, 'fix': fix}, index=symbols
        )

    def is_close_to_bound(self, values=None, zero_limit=0.01, significant_digits=2):
        """Logical Series of whether values are close to the respective bounds

        Parameters
        ----------
        values : pd.Series
            Series of values with index a subset of parameter names.
            Default is to use all parameter inits

        Returns
        -------
        pd.Series
            Logical Series with same index as values

        Example
        -------
        >>> from pharmpy import Parameters, Parameter
        >>> par1 = Parameter("CL", 1, lower=0, upper=10)
        >>> par2 = Parameter("V", 10, lower=0, upper=100)
        >>> pset = Parameters([par1, par2])
        >>> pset.is_close_to_bound()
        CL    False
        V     False
        dtype: bool
        """
        if values is None:
            values = pd.Series(self.inits)
        ser = pd.Series(
            [
                self[p].is_close_to_bound(values.loc[p], zero_limit, significant_digits)
                for p in values.index
            ],
            index=values.index,
            dtype=bool,
        )
        return ser

    @property
    def names(self):
        """List of all parameter names"""
        return [p.name for p in self._params]

    @property
    def symbols(self):
        """List of all parameter symbols"""
        return [p.symbol for p in self._params]

    @property
    def lower(self):
        """Lower bounds of all parameters as a dictionary"""
        return {p.name: p.lower for p in self._params}

    @property
    def upper(self):
        """Upper bounds of all parameters as a dictionary"""
        return {p.name: p.upper for p in self._params}

    @property
    def inits(self):
        """Initial estimates of parameters as dict"""
        return {p.name: p.init for p in self._params}

    @property
    def nonfixed_inits(self):
        """Dict of initial estimates for all non-fixed parameters"""
        return {p.name: p.init for p in self._params if not p.fix}

    @inits.setter
    def inits(self, init_dict):
        for key, _ in init_dict.items():
            if key not in self:
                raise KeyError(f'Parameter {key} not in Parameters')
        for name, value in init_dict.items():
            self[name].init = value

    @property
    def fix(self):
        """Fixedness of parameters as dict"""
        return {p.name: p.fix for p in self._params}

    @fix.setter
    def fix(self, fix_dict):
        for key in fix_dict:
            if key not in self:
                raise KeyError(f'Parameter {key} not in Parameters')
        for name, value in fix_dict.items():
            self[name].fix = value

    def remove_fixed(self):
        """Remove all fixed parameters"""
        nonfixed = [p for p in self._params if not p.fix]
        self._params = nonfixed

    def copy(self):
        """Create a deep copy of this Parameters"""
        return copy.deepcopy(self)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for p1, p2 in zip(self, other):
            if p1 != p2:
                return False
        return True

    def __repr__(self):
        if len(self) == 0:
            return "Parameters()"
        return self.to_dataframe().to_string()

    def _repr_html_(self):
        if len(self) == 0:
            return "Parameters()"
        else:
            return self.to_dataframe().to_html()

    def simplify(self, expr):
        """Simplify expression given constraints of parameters"""
        d = dict()
        for p in self._params:
            if p.fix:
                s = sympy.Float(p.init)
            elif p.upper < 0:
                s = sympy.Symbol(p.name, real=True, negative=True)
                d[s] = p.symbol
            elif p.upper <= 0:
                s = sympy.Symbol(p.name, real=True, nonpositive=True)
                d[s] = p.symbol
            elif p.lower > 0:
                s = sympy.Symbol(p.name, real=True, positive=True)
                d[s] = p.symbol
            elif p.lower >= 0:
                s = sympy.Symbol(p.name, real=True, nonnegative=True)
                d[s] = p.symbol
            else:
                s = sympy.Symbol(p.name, real=True)
                d[s] = p.symbol
            expr = expr.subs(p.symbol, s)
        simp = sympy.simplify(expr)  # Subs symbols back to non-constrained
        return simp.subs(d)


class Parameter:
    """A single parameter

    Example
    -------

    >>> from pharmpy import Parameter
    >>> param = Parameter("TVCL", 0.005, lower=0)
    >>> param.fix = True

    Parameters
    ----------
    name : str
        Name of the parameter
    init : number
        Initial estimate or simply the value of parameter.
    fix : bool
        A boolean to indicate whether the parameter is fixed or not. Note that fixing a parameter
        will keep its bounds even if a fixed parameter is actually constrained to one single
        value. This is so that unfixing will take back the previous bounds.
    lower : number
        The lower bound of the parameter. Default no bound. Must be less than the init.
    upper : number
        The upper bound of the parameter. Default no bound. Must be greater than the init.
    unit : str or unit expression
        Unit of parameter. Default to unitless
    """

    def __init__(self, name, init, lower=None, upper=None, fix=False, unit=1):
        self._init = init
        self.name = name
        self.fix = bool(fix)
        self._lower = -sympy.oo
        self._upper = sympy.oo
        if lower is not None:
            self.lower = lower
        if upper is not None:
            self.upper = upper
        self.unit = parse_units(unit)

    @property
    def symbol(self):
        """Symbol representing the parameter"""
        return symbols.symbol(self.name)

    @property
    def lower(self):
        """Lower bound of the parameter"""
        return self._lower

    @lower.setter
    def lower(self, new_lower):
        if new_lower > self.init:
            raise ValueError(f'Lower bound {new_lower} cannot be greater than init {self.init}')
        self._lower = new_lower

    @property
    def upper(self):
        """Upper bound of the parameter"""
        return self._upper

    @upper.setter
    def upper(self, new_upper):
        if new_upper < self.init:
            raise ValueError(f'Upper bound {new_upper} cannot be less than init {self.init}')
        self._upper = new_upper

    @property
    def init(self):
        """Initial parameter estimate or value"""
        return self._init

    @init.setter
    def init(self, new_init):
        if new_init < self.lower or new_init > self.upper:
            raise ValueError(
                f'Initial estimate must be within the constraints of the parameter: '
                f'{new_init} ∉ {sympy.pretty(sympy.Interval(self.lower, self.upper))}'
                f'\nUnconstrain the parameter before setting an initial estimate.'
            )
        self._init = new_init

    def is_close_to_bound(self, value=None, zero_limit=0.01, significant_digits=2):
        """Check if parameter value is close to any bound

        Parameters
        ----------
        value : number
            value to check against parameter bounds. Defaults to checking against the parameter init
        zero_limit : number
            maximum distance to 0 bounds
        significant_digits : int
            maximum distance to non-zero bounds in number of significant digits

        Examples
        --------

        >>> from pharmpy import Parameter
        >>> par = Parameter("x", 1, lower=0, upper=10)
        >>> par.is_close_to_bound()
        False

        >>> par.is_close_to_bound(0.005)
        True

        >>> par.is_close_to_bound(0.005, zero_limit=0.0001)
        False

        >>> par.is_close_to_bound(9.99)
        True

        >>> par.is_close_to_bound(9.99, significant_digits=3)
        False
        """
        if value is None:
            value = self.init
        return (
            self.lower > -sympy.oo
            and is_near_target(value, self.lower, zero_limit, significant_digits)
        ) or (
            self.upper < sympy.oo
            and is_near_target(value, self.upper, zero_limit, significant_digits)
        )

    def unconstrain(self):
        """Remove all constraints from this parameter

        Example
        -------
        >>> from pharmpy import Parameter
        >>> par = Parameter("x", 1, lower=0, upper=2)
        >>> par.unconstrain()
        >>> par
        Parameter("x", 1, lower=-oo, upper=oo, fix=False)
        """
        self._lower = -sympy.oo
        self._upper = sympy.oo
        self.fix = False

    @property
    def parameter_space(self):
        """The parameter space set

        A fixed parameter will be constrained to one single value
        and non-fixed parameters will be constrained to an interval
        possibly open in one or both ends.

        Examples
        --------

        >>> import sympy
        >>> from pharmpy import Parameter
        >>> par = Parameter("x", 1, lower=0, upper=10)
        >>> sympy.pprint(par.parameter_space)
        [0, 10]

        >>> par.fix = True
        >>> sympy.pprint(par.parameter_space)
        {1}
        """
        if self.fix:
            return sympy.FiniteSet(self.init)
        else:
            return sympy.Interval(self.lower, self.upper)

    def copy(self):
        """Create a copy of this Parameter"""
        return copy.deepcopy(self)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        """Two parameters are equal if they have the same name, init and constraints"""
        return (
            self.init == other.init
            and self.lower == other.lower
            and self.upper == other.upper
            and self.name == other.name
            and self.fix == other.fix
        )

    def __repr__(self):
        return (
            f'Parameter("{self.name}", {self.init}, lower={self.lower}, upper={self.upper}, '
            f'fix={self.fix})'
        )



import pytest
import sympy

from pharmpy.config import ConfigurationContext
from pharmpy.model import ModelSyntaxError
from pharmpy.plugins.nonmem import conf
from pharmpy.symbols import symbol


def S(x):
    return symbol(x)


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,comment,results',
    [
        ('$OMEGA 1', False, [('OMEGA(1,1)', 1, 0, sympy.oo, False)]),
        ('$OMEGA   0.123 \n\n', False, [('OMEGA(1,1)', 0.123, 0, sympy.oo, False)]),
        ('$OMEGA   (0 FIX) ; CL', False, [('OMEGA(1,1)', 0, 0, sympy.oo, True)]),
        (
            '$OMEGA DIAG(2) 1 2 FIX',
            False,
            [
                ('OMEGA(1,1)', 1, 0, sympy.oo, False),
                ('OMEGA(2,2)', 2, 0, sympy.oo, True),
            ],
        ),
        (
            '$OMEGA 1 2 3',
            False,
            [
                ('OMEGA(1,1)', 1, 0, sympy.oo, False),
                ('OMEGA(2,2)', 2, 0, sympy.oo, False),
                ('OMEGA(3,3)', 3, 0, sympy.oo, False),
            ],
        ),
        ('$OMEGA 0.15 ;CL', False, [('OMEGA(1,1)', 0.15, 0, sympy.oo, False)]),
        (
            '$OMEGA 1 \n2 ; S   \n  3 ',
            False,
            [
                ('OMEGA(1,1)', 1, 0, sympy.oo, False),
                ('OMEGA(2,2)', 2, 0, sympy.oo, False),
                ('OMEGA(3,3)', 3, 0, sympy.oo, False),
            ],
        ),
        ('$OMEGA 2 SD', False, [('OMEGA(1,1)', 4, 0, sympy.oo, False)]),
        ('$OMEGA ;CO\n (VAR 3)', False, [('OMEGA(1,1)', 3, 0, sympy.oo, False)]),
        (
            '$OMEGA (1)x2',
            False,
            [
                ('OMEGA(1,1)', 1, 0, sympy.oo, False),
                ('OMEGA(2,2)', 1, 0, sympy.oo, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) 1 0.5 2',
            False,
            [
                ('OMEGA(1,1)', 1, 0, sympy.oo, False),
                ('OMEGA(2,1)', 0.5, -sympy.oo, sympy.oo, False),
                ('OMEGA(2,2)', 2, 0, sympy.oo, False),
            ],
        ),
        ('$OMEGA BLOCK(2) SAME', False, []),
        ('$OMEGA BLOCK SAME(3)', False, []),
        (
            '$OMEGA BLOCK(2) 1 (0.1)x2',
            False,
            [
                ('OMEGA(1,1)', 1, 0, sympy.oo, False),
                ('OMEGA(2,1)', 0.1, -sympy.oo, sympy.oo, False),
                ('OMEGA(2,2)', 0.1, 0, sympy.oo, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) CHOLESKY 0.8 -0.3 0.7',
            False,
            [
                ('OMEGA(1,1)', 0.64, 0, sympy.oo, False),
                ('OMEGA(2,1)', -0.24, -sympy.oo, sympy.oo, False),
                ('OMEGA(2,2)', 0.58, 0, sympy.oo, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) SD 0.8 -0.394 0.762 CORR',
            False,
            [
                ('OMEGA(1,1)', 0.64, 0, sympy.oo, False),
                ('OMEGA(2,1)', -0.2401824, -sympy.oo, sympy.oo, False),
                ('OMEGA(2,2)', 0.580644, 0, sympy.oo, False),
            ],
        ),
        (
            '$OMEGA BLOCK(1)   1.5',
            False,
            [
                ('OMEGA(1,1)', 1.5, 0, sympy.oo, False),
            ],
        ),
        (
            '$OMEGA  0.0258583  ;      V2\n'
            ';$OMEGA BLOCK(1) 0.0075 FIX    ;.02 ; IOC\n'
            ';$OMEGA BLOCK(1) SAME\n',
            False,
            [
                ('OMEGA(1,1)', 0.0258583, 0, sympy.oo, False),
            ],
        ),
        ('$OMEGA 1 ; IVCL', True, [('IVCL', 1, 0, sympy.oo, False)]),
        (
            '$OMEGA DIAG(2) 1 ; V1 df\n 2 FIX ; VA2 __12\n',
            True,
            [
                ('V1', 1, 0, sympy.oo, False),
                ('VA2', 2, 0, sympy.oo, True),
            ],
        ),
        (
            '$OMEGA BLOCK(2) 1 ;IV1\n 2 ;CORR\n 3 ;IV2',
            True,
            [
                ('IV1', 1, 0, sympy.oo, False),
                ('CORR', 2, -sympy.oo, sympy.oo, False),
                ('IV2', 3, 0, sympy.oo, False),
            ],
        ),
    ],
)
def test_parameters(parser, buf, comment, results):
    if comment:
        opt = ['comment', 'basic']
    else:
        opt = ['basic']
    with ConfigurationContext(conf, parameter_names=opt):
        recs = parser.parse(buf)
        rec = recs.records[0]
        pset, _, _ = rec.parameters(1, 1)
        assert len(pset) == len(results)
        for res in results:
            name = res[0]
            init = res[1]
            lower = res[2]
            upper = res[3]
            fix = res[4]
            param = pset[name]
            assert param.name == name
            assert pytest.approx(param.init, 0.00000000000001) == init
            assert param.lower == lower
            assert param.upper == upper
            assert param.fix == fix


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf, exc_msg',
    [
        ('$OMEGA 0 ', 'If initial estimate for OMEGA is'),
        ('$OMEGA DIAG(1) 1 SD VARIANCE', 'Initial estimate for OMEGA cannot'),
        ('$OMEGA SD BLOCK(2) 0.1 0.001 0.1 STANDARD', 'Cannot specify either option'),
        ('$OMEGA CHOLESKY BLOCK(2) 0.1 VAR 0.001 \n  0.1 ', 'Cannot specify either option'),
        ('$OMEGA BLOCK(3) 0.1 \n  0.1', 'Wrong number of inits in BLOCK'),
    ],
)
def test_errors(parser, buf, exc_msg):
    recs = parser.parse(buf)
    rec = recs.records[0]
    with pytest.raises(ModelSyntaxError, match=exc_msg):
        pset, _, _ = rec.parameters(1, None)


def test_parameters_offseted(parser):
    rec = parser.parse("$OMEGA 1").records[0]
    pset, _, _ = rec.parameters(3, None)
    assert pset['OMEGA(3,3)'].init == 1

    rec = parser.parse('$OMEGA   BLOCK   SAME').records[0]
    pset, _, _ = rec.parameters(3, 2)
    assert len(pset) == 0


def test_update(parser):
    rec = parser.parse('$OMEGA 1').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 2
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2'

    rec = parser.parse('$OMEGA 1 SD').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 SD'

    rec = parser.parse('$OMEGA (1)x3\n;FTOL').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,2)'].init = 4
    pset['OMEGA(3,3)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA (4)x3\n;FTOL'

    rec = parser.parse('$OMEGA (1)x2 2').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 1
    pset['OMEGA(2,2)'].init = 2
    pset['OMEGA(3,3)'].init = 0.5
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA (1) (2) 0.5'

    rec = parser.parse("$OMEGA DIAG(2) (1 VAR) (SD 2)").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 1.5
    pset['OMEGA(2,2)'].init = 16
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA DIAG(2) (1.5 VAR) (SD 4)'

    rec = parser.parse("$OMEGA BLOCK(2) 1 2 4").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 7
    pset['OMEGA(2,1)'].init = 0.5
    pset['OMEGA(2,2)'].init = 8
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 7 0.5 8'

    rec = parser.parse("$OMEGA BLOCK(2)\n SD 1 0.5 ;COM \n 1\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,1)'].init = 0.25
    pset['OMEGA(2,2)'].init = 9
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2)\n SD 2 0.25 ;COM \n 3\n'

    rec = parser.parse("$OMEGA CORR BLOCK(2)  1 0.5 1\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,1)'].init = 1
    pset['OMEGA(2,2)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA CORR BLOCK(2)  4 0.25 4\n'

    rec = parser.parse("$OMEGA CORR BLOCK(2)  1 0.5 SD 1\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,1)'].init = 1
    pset['OMEGA(2,2)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA CORR BLOCK(2)  2 0.25 SD 2\n'

    rec = parser.parse("$OMEGA BLOCK(2) 1 0.1 1\n CHOLESKY").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 0.64
    pset['OMEGA(2,1)'].init = -0.24
    pset['OMEGA(2,2)'].init = 0.58
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 0.8 -0.3 0.7\n CHOLESKY'

    rec = parser.parse("$OMEGA BLOCK(2) (1)x3\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 0.64
    pset['OMEGA(2,1)'].init = -0.24
    pset['OMEGA(2,2)'].init = 0.58
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 0.64 -0.24 0.58\n'

    rec = parser.parse("$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 1
    pset['OMEGA(2,1)'].init = 0.2
    pset['OMEGA(2,2)'].init = 3
    pset['OMEGA(3,1)'].init = 0.4
    pset['OMEGA(3,2)'].init = 0.5
    pset['OMEGA(3,3)'].init = 6
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(3) 1 ;CL\n0.2 3; V\n0.4 0.5 6; KA'

    rec = parser.parse("$OMEGA BLOCK(2) SAME").records[0]
    pset, _, _ = rec.parameters(3, 2)
    assert len(pset) == 0
    next_eta, prev_size = rec.update(pset, 3, 2)
    assert next_eta == 5
    assert prev_size == 2
    assert str(rec) == "$OMEGA BLOCK(2) SAME"

    rec = parser.parse("$OMEGA BLOCK SAME").records[0]
    pset, _, _ = rec.parameters(5, 4)
    assert len(pset) == 0
    next_eta, prev_size = rec.update(pset, 5, 4)
    assert next_eta == 9
    assert prev_size == 4
    assert str(rec) == "$OMEGA BLOCK SAME"


def test_update_fix(parser):
    rec = parser.parse("$OMEGA 2 FIX 4 (FIX 6)").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = False
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 4 (FIX 6)'

    rec = parser.parse("$OMEGA 2 4 6 ;STRAML").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = True
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 FIX 4 6 ;STRAML'
    pset['OMEGA(1,1)'].fix = False
    pset['OMEGA(2,2)'].fix = True
    pset['OMEGA(3,3)'].fix = True
    pset['OMEGA(3,3)'].init = 23
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 4 FIX 23 FIX ;STRAML'

    rec = parser.parse("$OMEGA BLOCK(2) 1 .1 2 ;CLERMT").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = True
    # Cannot fix parts of block
    with pytest.raises(ValueError):
        rec.update(pset, 1, None)
    pset['OMEGA(2,1)'].fix = True
    pset['OMEGA(2,2)'].fix = True
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) FIX 1 .1 2 ;CLERMT'

    rec = parser.parse("$OMEGA BLOCK(2) 1 .1 FIX 1").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = False
    pset['OMEGA(2,1)'].fix = False
    pset['OMEGA(2,2)'].fix = False
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 1 .1 1'


def test_random_variables(parser):
    rec = parser.parse("$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA").records[0]
    rvs, nxt, cov, zero_fix = rec.random_variables(1)
    assert nxt == 4
    assert len(rvs) == 3
    assert rvs[0].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert rvs[2].name == 'ETA(3)'
    assert len(cov) == 9
    assert len(zero_fix) == 0

    rec = parser.parse("$OMEGA BLOCK(1) 1.5").records[0]
    rvs, nxt, cov, zero_fix = rec.random_variables(2)
    assert nxt == 3
    assert len(rvs) == 1
    assert rvs[0].name == 'ETA(2)'
    assert isinstance(rvs[0].sympy_rv.pspace.distribution, sympy.stats.crv_types.NormalDistribution)
    assert cov == S('OMEGA(2,2)')
    assert len(zero_fix) == 0

    p = parser.parse("$OMEGA BLOCK(2) 1 0.01 1\n$OMEGA BLOCK(2) SAME\n")
    rec0 = p.records[0]
    rec1 = p.records[1]
    rvs, nxt, cov, zero_fix = rec0.random_variables(1)
    assert nxt == 3
    assert len(rvs) == 2
    assert rvs[0].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert len(cov) == 4
    assert len(zero_fix) == 0
    A = sympy.Matrix([[S('OMEGA(1,1)'), S('OMEGA(2,1)')], [S('OMEGA(2,1)'), S('OMEGA(2,2)')]])
    assert rvs[0].sympy_rv.pspace.distribution.sigma == A
    rvs, nxt, cov, zero_fix = rec1.random_variables(nxt, cov)
    assert nxt == 5
    assert len(rvs) == 2
    assert rvs[0].name == 'ETA(3)'
    assert rvs[1].name == 'ETA(4)'
    assert len(cov) == 4
    assert len(zero_fix) == 0
    assert rvs[0].sympy_rv.pspace.distribution.sigma == A

    rec = parser.parse("$OMEGA 0 FIX").records[0]
    rvs, _, _, zero_fix = rec.random_variables(1)
    assert len(rvs) == 0
    assert zero_fix == ['ETA(1)']

    rec = parser.parse("$OMEGA  BLOCK(2) FIX 0 0 0").records[0]
    rvs, _, _, zero_fix = rec.random_variables(1)
    assert len(rvs) == 0
    assert zero_fix == ['ETA(1)', 'ETA(2)']

    p = parser.parse("$OMEGA BLOCK(1) 0 FIX\n$OMEGA BLOCK(1) SAME")
    rec0 = p.records[0]
    rec1 = p.records[1]
    rvs, nxt, _, zero_fix = rec0.random_variables(1)
    assert nxt == 2
    assert len(rvs) == 0
    assert zero_fix == ['ETA(1)']
    rvs, nxt, _, zero_fix = rec1.random_variables(2, previous_cov='ZERO')
    assert nxt == 3
    assert len(rvs) == 0
    assert zero_fix == ['ETA(2)']

    rec = parser.parse("$OMEGA BLOCK SAME").records[0]
    A = sympy.Matrix([[1, 0.01], [0.01, 1]])
    rvs, _, _, _ = rec.random_variables(3, A)
    assert len(rvs) == 2
    assert list(rvs)[0].name == 'ETA(3)'
    assert list(rvs)[1].name == 'ETA(4)'

    with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
        p = parser.parse("$OMEGA BLOCK(2) 1 ;IV1\n 0.01 ;CORR\n 1 ;IV2\n$OMEGA BLOCK(2) SAME\n")
        rec0 = p.records[0]
        rec1 = p.records[1]
        rvs, nxt, cov, zero_fix = rec0.random_variables(1)
        assert nxt == 3
        assert len(rvs) == 2
        assert rvs[0].name == 'ETA(1)'
        assert rvs[1].name == 'ETA(2)'
        assert len(cov) == 4
        assert len(zero_fix) == 0
        A = sympy.Matrix([[S('IV1'), S('CORR')], [S('CORR'), S('IV2')]])
        assert rvs[0].sympy_rv.pspace.distribution.sigma == A
        rvs, nxt, cov, zero_fix = rec1.random_variables(nxt, cov)
        assert nxt == 5
        assert len(rvs) == 2
        assert rvs[0].name == 'ETA(3)'
        assert rvs[1].name == 'ETA(4)'
        assert len(cov) == 4
        assert len(zero_fix) == 0
        assert rvs[0].sympy_rv.pspace.distribution.sigma == A


@pytest.mark.parametrize(
    'buf,remove,result',
    [
        (
            '$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA',
            {'ETA(1)'},
            '$OMEGA BLOCK(2)\n1.0\n0.1 1.0\n',
        ),
        (
            '$OMEGA BLOCK(4) 1 ;CL\n0.2 3; V\n0.4 0.5 6; KA\n7 8 9 10\n',
            {'ETA(2)', 'ETA(4)'},
            '$OMEGA BLOCK(2)\n1.0\n0.4 6.0\n',
        ),
        ('$OMEGA 1 2 3\n', {'ETA(2)'}, '$OMEGA 1  3\n'),
        ('$OMEGA BLOCK(2) 1 2 3 FIX\n', {'ETA(1)'}, '$OMEGA BLOCK(1) FIX\n3.0\n'),
    ],
)
def test_remove(parser, buf, remove, result):
    rec = parser.parse(buf).records[0]
    rec.random_variables(1)
    rec.remove(remove)
    assert str(rec) == result


def test_iov(parser):
    rec = parser.parse('$OMEGA BLOCK(2) SAME').records[0]
    rvs, _, _, _ = rec.random_variables(1, sympy.Matrix([[2, 0], [0, 1]]))
    assert rvs[0].level == 'IOV'
    assert rvs[1].level == 'IOV'

    rec = parser.parse('$OMEGA 1').records[0]
    rvs, _, _, _ = rec.random_variables(1)
    assert rvs[0].level == 'IIV'



from io import StringIO

from pyfakefs.fake_filesystem_unittest import Patcher

from pharmpy import Model
from pharmpy.modeling import (
    has_additive_error_model,
    has_combined_error_model,
    has_proportional_error_model,
    load_example_model,
    read_model_from_string,
    remove_error_model,
    set_additive_error_model,
    set_combined_error_model,
    set_dtbs_error_model,
    set_proportional_error_model,
    set_weighted_error_model,
    use_thetas_for_error_stdev,
)
from pharmpy.modeling.error import _get_prop_init
from pharmpy.statements import Assignment


def test_remove_error_model(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    remove_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y = F'


def test_set_additive_error_model(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    set_additive_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y = F + EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA  11.2225 ; sigma'
    before = model.model_code
    set_additive_error_model(model)  # One more time and nothing should change
    assert before == model.model_code


def test_set_additive_error_model_logdv(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    set_additive_error_model(model, data_trans="log(Y)")
    assert model.model_code.split('\n')[11] == 'Y = LOG(F) + EPS(1)/F'
    assert model.model_code.split('\n')[17] == '$SIGMA  11.2225 ; sigma'


def test_set_proportional_error_model(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    model.statements[5] = Assignment('Y', 'F')
    set_proportional_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y=F+F*EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA  0.09 ; sigma'

    model = Model(testdata / 'nonmem' / 'pheno.mod')
    set_proportional_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y=F+F*EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA 0.013241'


def test_set_proportional_error_model_log(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    model.statements[5] = Assignment('Y', 'F')
    set_proportional_error_model(model, data_trans='log(Y)')
    assert model.model_code.split('\n')[11] == 'Y = LOG(F) + EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA  0.09 ; sigma'


def test_set_combined_error_model(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    set_combined_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y = F + EPS(1)*F + EPS(2)'
    assert model.model_code.split('\n')[17] == '$SIGMA  0.09 ; sigma_prop'
    assert model.model_code.split('\n')[18] == '$SIGMA  11.2225 ; sigma_add'
    before = model.model_code
    set_combined_error_model(model)  # One more time and nothing should change
    assert before == model.model_code


def test_set_combined_error_model_log(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    set_combined_error_model(model, data_trans='log(Y)')
    assert model.model_code.split('\n')[11] == 'Y = LOG(F) + EPS(2)/F + EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA  0.09 ; sigma_prop'
    assert model.model_code.split('\n')[18] == '$SIGMA  11.2225 ; sigma_add'


def test_remove_error_without_f(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = Model(StringIO(code))
    remove_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
Y = CONC
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_additive_error_without_f(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = Model(StringIO(code))
    model.dataset = load_example_model("pheno").dataset
    set_additive_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y = CONC + EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA  11.2225 ; sigma
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_get_prop_init(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')

    init = _get_prop_init(model.dataset)
    assert init == 11.2225

    model.dataset['DV'].values[:] = 0.0
    init = _get_prop_init(model.dataset)
    assert init == 0.01


def test_has_additive_error_model(testdata):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_additive_error_model(model)
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_additive_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = CONC + EPS(1) + CONC*EPS(2)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_additive_error_model(model)

    model = Model(StringIO(code))
    assert not has_additive_error_model(model)


def test_has_proportional_error_model(testdata):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_proportional_error_model(model)
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_proportional_error_model(model)

    model = Model(StringIO(code))
    assert has_proportional_error_model(model)


def test_has_combined_error_model(testdata):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_combined_error_model(model)
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_combined_error_model(model)

    model = Model(StringIO(code))
    assert not has_combined_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1) + ERR(2)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA .02
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_combined_error_model(model)


def test_use_theta_for_error_stdev():
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    use_thetas_for_error_stdev(model)
    correct = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)*THETA(2)

$THETA 0.1
$THETA  (0,1.0) ; SD_EPS(1)
$OMEGA 0.01
$SIGMA 1 FIX
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""

    assert model.model_code == correct


def test_set_weighted_error_model():
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    set_weighted_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
W = CONC
Y = CONC + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""

    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*THETA(3)*EPS(1)+THETA(4)*EPS(2)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA (0, 0.1) ; SD1
$THETA (0, 0.2) ; SD2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    set_weighted_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
W = SQRT(CONC**2*THETA(3)**2 + THETA(4)**2)
Y = CONC + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA (0, 0.1) ; SD1
$THETA (0, 0.2) ; SD2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_set_dtbs_error_model():
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
Y=CONC+CONC*EPS(1)+EPS(2)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model.name = 'run1'
    set_dtbs_error_model(model)

    with Patcher(additional_skip_names=['pkgutil']):
        model.update_source()
        with open('run1_contr.f90') as fh:
            assert fh.readline().startswith('      subroutine contr')
        with open('run1_ccontra.f90') as fh:
            assert fh.readline().startswith('      subroutine ccontr')

    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
W = SQRT(CONC**2*THETA(3)**2 + THETA(4)**2)
W = CONC**THETA(6)*W
IF (CONC.NE.0.AND.THETA(5).NE.0) THEN
IPRED = (CONC**THETA(5) - 1)/THETA(5)
ELSE IF (THETA(5).EQ.0.AND.CONC.NE.0) THEN
IPRED = LOG(CONC)
ELSE IF (CONC.EQ.0.AND.THETA(5).EQ.0) THEN
IPRED = -1/THETA(5)
ELSE
IPRED = -1000000000
END IF
Y = IPRED + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA  (0,0.11506954418958998) ; SD_EPS(1)
$THETA  (0,0.11506954418958998) ; SD_EPS(2)
$THETA  1 ; tbs_lambda
$THETA  0.001 ; tbs_zeta
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 1 FIX
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct
